// Comp 2673, Spring 2003
// Incomplete example, demonstrates dynamically allocated memory in a class

// class for storing edges in a graph, weighted graph, or directed graph
// Under normal circumstances, this definition would be stored in a header
// file graph.h, and the implementation of these functions would be in a
// separate .cpp file
// Note that this class does not dynamically allocate memory, so the
// copy constructor and assigment operator should just do a memberwise
// copy.  The destructor needs only reset the variables values to 0.
class edge {
public:
   edge();                        // default constructor
   edge(int sv, int ev, float w); // initializing constructor
   edge(const edge &orig);        // copy constructor
   virtual ~edge();               // destructor 
   int startv() const;            // returns the starting vertex (get function)
   int endv() const;              // returns the ending vertex (get function)
   float weight() const;          // returns the weight (get function)
   const edge & operator=(const edge &rhs);    // assignment operator
   bool set_values(int sv, int ev, float wt);  // set all values
   bool startv(int v);                         // set the start vertex
   bool endv(int v);                           // set the end vertex
   bool weight(float w);                       // set the weight
   void display(ostream &o) const; // display function can output to any ostream
private:
   int startv;
   int endv;
   float weight;
};

// class for storing data for a graph.  Again, this definition should go
// in a header file, and the implementation of member functions belongs
// in a separate .cpp file.  This class dynamically allocates memory in the
// edgelist.  So any time that the edgelist is changed, you must remember
// to deallocate and reallocate memory.  In particular, this is done in
// all of the functions shown.
class graph {
public:
    graph();                           // default constructor
    graph(const graph &orig);          // copy constructor
    bool read_graph(const string &filename);    // read graph data out of file
    const graph & operator=(const graph &rhs);  // assignment operator
    virtual ~graph();                           // destructor
    void display(ostream &o) const; // display function can outpu to any ostream
    ... and other useful functions
private:
    int num_vertices;
    int num_edges;
    edge *edgelist;      // this pointer will point to a dyn. allocated array of
                         // edges.   The graph object owns this memory.
};

// stream insertion operator, allows graph to be output to a file or to
// cout with a line like:
//  cout << graph1 << graph2;
// The prototype for this belongs in the file graph.h
ostream & operator<<(ostream &o, const graph &g)
{
    g.display(o);
    return o;
}

// Here's a function that shows examples of creating objects of the above
// classes
int main()
{
    edge edge1;              // statically allocates an edge.  Default
                             // constructor is called.
    edge1.set_values(2, 4, 6.8);  // alter values in edge1
    edge edge2(2, 3, 47.6);  // Another statically allocated edge.  Initializing
                             // constructor is called.
    edge *ep;                // pointer to an edge
    ep = new edge(1, 2, 3);  // space for an edge is created, constructor called
    ep->startv(3);           // sets the start vertex to 3
    cout << ep->startv();    // accesses the startv, and outputs it
    delete ep;               // deallocates space and calls destructor
    graph g;                 // statically allocates g, calls def. constructor
    g.read_graph("foo.grf"); // read in the graph
    graph g2(g);             // copy constructor called

    g2.display(cout);        // These two lines show two ways that you
    cout << g2 << endl;      // can output a graph

    ofstream f("foo.grf");   // We overloaded the output operator in a flexible
    f << g2 << endl;         // way, so we can now use it to output to files

    graph *gp;               // pointer to a graph;
    gp = new graph;          // allocates space for graph, def. const. called
                             // no space allocated for edges
    *gp = g2;                // assignment operator called

    cout << *gp << endl;     // This shows two ways to display a graph
    gp->display(cout);       // via a pointer.

    delete gp;               // de-allocates space, destructor called
   
}  // Statically allocated automatic variables are released at the end of the
   // block they're defined in.  edge1, edge2, g, and g2 are released on this
   // line and the appropriate destructor is called 

// Below is the implementation of some of the trickier functions in
// the graph class.  Under normal circumstances, all code below would
// go in the file graph.cpp

graph::graph()
  // initialize values to 0
  : num_vertices(0), num_edges(0), data(0)
{
}

// copy constructor for the graph class
graph::graph(const graph &orig)
  // copy number of vertices and number of edges from original
  // set pointer to 0 for safety
  : num_vertices(orig.num_vertices), num_edges(orig.num_edges), data(0)
{
    edgelist = new edge[num_edges];       // Make space for the edgelist
    if (!edgelist) {
        cerr << "Graph:  Unable to allocate memory.  Exiting program...";
        exit(1);
    }
    // now do a deep copy of the edgelist
    for (int i = 0; i < num_edges; i++)   // Make a copy of every edge
        edgelist[i] = orig.edgelist[i];
}

// destructor
graph::~graph()
{
   // delete any memory that is owned by this object
   if (edgelist != 0)
       delete [] edgelist; 
   // reset all variables to 0 for safety
   edgelist = 0;
   num_vertices = 0;
   num_edges = 0;
}

// assignment operator
const graph & graph::operator=(const graph &rhs)
{
    // First make sure we're not copying from this object to itself
    if (this != &rhs) {
        // release an memory that is already there
        if (edgelist != 0) {
            delete [] edgelist;
            edgelist = 0;
        }
        // copy the number of vertices and number of edges
        num_vertices = rhs.num_vertices;
        num_edges = rhs.num_edges;
        // allocate space for edgelist
        edgelist = new edge[num_edges];       // Make space for the edgelist
        // make sure memory allocation succeeded.
        if (!edgelist) {
           cerr << "Graph:  Unable to allocate memory.  Exiting program...";
           exit(1);
        }
        // now do a deep copy of the edgelist
        for (int i = 0; i < num_edges; i++)   // Make a copy of every edge
            edgelist[i] = rhs.edgelist[i];
    }
    return *this;
}

// function for reading a graph from a file
bool graph::read_graph(const string &filename)
{
    // Insert code here to open the file, return false in case of error
    // Insert code here to read the number of vertices and number of edges,
    //  store their values in num_vertices and num_edges
    // If there already is a graph in this object, then release the memory
    if (!edgelist)
       delete [ ] edgelist;
    // allocate space for the new edgelist
    edgelist = new edge[num_edges];
    // check that the allocation worked, exit if not
    if (!edgelist) {
        cerr << "Graph:  Unable to allocate memory.  Exiting program...";
        exit(1);
    }
    // insert code here to loop to read all the edges, store in edgelist
    //    if there's an error, release the memory, reset all variables, and
    //    return false
    // if all was successful, then return true:
    return true;
}

